package org.eso.ohs.core.utilities;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

/**
 * Utility methods to convert data between formats, etc.
 *
 * @author A M Chavan, 06-Jun-2001
 * @version $Revision: 1.20 $
 */

// $Id: Convert.java,v 1.20 2005/11/15 13:16:46 ddorigo Exp $

public class Convert {

//    private static Logger stdlog_ = Logger.getLogger(Convert.class);

    public final static String ST_TIME_FORMAT = "HH:mm";

    public final static int SEC_PER_MINUTE = 60;

    public final static int SEC_PER_HOUR = 3600;

    public final static int SEC_PER_DAY = 3600 * 24;

    /** Milliseconds per second */
    public final static int MILLISECS_PER_SEC = 1000;

    /** Milliseconds per minute */
    public final static int MILLISECS_PER_MINUTE = 60 * MILLISECS_PER_SEC;

    /** Milliseconds per unit (hour or degree) */
    public final static int MILLISECS_PER_UNIT = 60 * MILLISECS_PER_MINUTE;

    /** Number of decimal places of accuracy */
    public final static int DECIMAL_PLACES_ACCURACY = 3;

    /** Max milliseconds for RA -- corresponds to 23:59:59.999 */
    public final static int MAX_MILLISECS_RA = 86400000;

    /** Max milliseconds for DEC -- corresponds to 90:00:00.000 */
    public final static int MAX_MILLISECS_DEC = 324000000;

    /** Max execution time 23:59:59.999 */
    public final static int EXECTIME_MAX = 86399;
    public final static int EXECTIME_ERR = -86399;
    public final static String EXECTIME_ERR_STRING = "-23:59:59.000";

    /** Greatest date value "2038-01-19T02:14:07" allowed */
    public final static long TIME_MAX = 2147483648L;

    public final static long ENDTIME_MAX = 2147483647L;

    /** Standard format for ISO date/time strings.
     * @author amchavan, 03-Jun-2002: added
     */
    public final static String ISO_DATE_FORMAT = "yyyy-MM-dd'T'HH:mm:ss";

    /** Standard format for duration strings.
    *
    */
   public static final String DURATION_FORMAT = "d'T'HH:mm:ss";
   /** Standard format for duration strings.
   *
   */
  public static final String DURATION_HOURS = "HH:mm:ss";
  /** Standard format for duration strings.
   *
   */
  public static final String DURATION_DAYS = "d";

    /**  
     * @return Number of seconds since the Unix epoch.
     */

    public static long getCurrentTime() {
        return System.currentTimeMillis() / Convert.MILLISECS_PER_SEC;
    }

    /* 
    * Convert hours and minutes string to seconds.
    * @param hh_mm A string like <code>HH:MM</code>.
    */

    public static int convertHHMMTOSeconds(String hh_mm)
        throws NumberFormatException {
        int seconds = 0;
        String[] array = hh_mm.split(":");
        if (array == null || array.length != 2)
            throw new NumberFormatException(
                "Invalid format should be hh:mm not " + hh_mm);
        if(array[0].length() > 2 || array[1].length() > 2)        
            throw new NumberFormatException(
                "Invalid format should be hh:mm not " + hh_mm);                
        try {
            int hours = Integer.valueOf(array[0]).intValue();
            int minutes = Integer.valueOf(array[1]).intValue();
            seconds =
                hours * Convert.SEC_PER_HOUR + minutes * Convert.SEC_PER_MINUTE;
            return seconds;
        } catch (NumberFormatException e) {
//            stdlog_.debug("HH MM " + hh_mm);
            throw new NumberFormatException(
                "Invalid format should be hh:mm not " + hh_mm);
        } catch (IndexOutOfBoundsException e) {
            throw new NumberFormatException(
                "Invalid format should be hh:mm not " + hh_mm);
        }
    }

    /**
     * Converts a string in the format HH:MM:SS.sss to
     * milliarcseconds.
     *
     * @param value  value to be converted, represented 
     *               as <code>HH:MM:SS.sss</code> or  
     *                  <code>HHMMSS.sss</code>, 
     *               where HH is hours (0-23), 
     *                     MM is minutes (0-59), 
     *                     SS is seconds (0-59) and 
     *                     sss is milliseconds. <P>
     *
     *        <strong>Note</strong>: partial strings are accepted and
     *        filled in as follows: <BR>
     *
     *        if the input string includes a colon char (:), then
     *        missing elements are added to the <em>right</em> of it;
     *        e.g. <code>12:3</code> is equivalent to
     *        <code>12:03:00.000</code>;
     *        <BR>
     *
     *        if the input string does not include any colon chars,
     *        then missing elements are added to the <em>left</em> of it;
     *        e.g. <code>123</code> is equivalent to
     *        <code>00:01:23.000</code>.
     *
     * @return		input converted to milliarcseconds
     *
     * @exception       IllegalArgumentException on out of bounds and
     *                                              out of format
     *
     * @author A M Chavan, 06-Jun-2001; was originally
     *         convertToMilliArcSeconds() in DegreeAdaptor.java
     */
    public static int HHMMSSToMilliarcsec(String value)
        throws IllegalArgumentException {
        return (stringToMilliarcsec(value, true, false));
    }

    /**
     * Converts a string in the format HH:MM:SS.sss to
     * seconds. Calls stringToMilliarcsec and divides the
     * return value by 1000 (millisecs per sec). 
     *
     * @param value  value to be converted, represented
     *			as <code>HH:MM:SS.sss</code> or
     *			<code>HHMMSS.sss</code>,
     *			where HH is hours (0-23),
     *			   MM is minutes (0-59),
     *			   SS is seconds (0-59) and
     *			   sss is milliseconds. <P>
     *
     *		<strong>Note</strong>: partial strings are accepted and
     *		filled in as follows: <BR>
     *
     *		if the input string includes a colon char (:), then
     *		missing elements are added to the <em>right</em> of it;
     *		e.g. <code>12:3</code> is equivalent to
     *		<code>12:03:00.000</code>;
     *		<BR>
     *
     *		if the input string does not include any colon chars,
     *		then missing elements are added to the <em>left</em> of it;
     *		e.g. <code>123</code> is equivalent to
     *		<code>00:01:23.000</code>.
     *
     * @return		input converted to seconds
     *
     * @exception	IllegalArgumentException on out of bounds and
     *                                              out of format
     *
     * @author cmavros, 10-10-01
     */
    public static int HHMMSSToSec(String value)
        throws IllegalArgumentException {
        int newValue = stringToMilliarcsec(value, false, true);
        return (newValue / MILLISECS_PER_SEC);
    }

    /**
     * Converts a Declination string to milliarcseconds, 
     *
     * @param value  value to be converted, represented 
     *               as <code>[-]DD:MM:SS.sss</code> or  
     *                  <code>[-]DDMMSS.sss</code>, 
     *               where DD is degrees (0-90), 
     *                     MM is minutes (0-59), 
     *                     SS is seconds (0-59) and 
     *                     sss is milliseconds. <P>
     *
     *        <strong>Note</strong>: partial strings are are accepted
     *        and filled in as follows: <BR>
     *
     *        if the input string includes a colon char (:), then
     *        missing elements are added to the <em>right</em> of it;
     *        e.g. <code>12:3</code> is equivalent to
     *        <code>12:03:00.000</code>;
     *        <BR>
     *
     *        if the input string does not include any colon chars,
     *        then missing elements are added to the <em>left</em> of it;
     *        e.g. <code>123</code> is equivalent to
     *        <code>00:01:23.000</code>.
     *
     * @return		input converted to milliarcseconds
     *
     * @exception       IllegalArgumentException on out of bounds and
     *                                              out of format
     *
     * @author A M Chavan, 06-Jun-2001; was originally
     *         convertToMilliArcSeconds() in DegreeAdaptor.java
     */
    public static int DDMMSSToMilliarcsec(String value)
        throws IllegalArgumentException {

        return (stringToMilliarcsec(value, false, false));
    }

    /**
     * Converts a Declination string to milliarcseconds,
     *
     * @param value  value to be converted
     * @param noCheck  boolean used to turn off bounds checking
     *
     * @return		input converted to milliarcseconds
     * @exception       IllegalArgumentException on out of bounds and
     *                  out of format
     * @see #DDMMSSToMilliarcsec(String value)
     * @deprecated Use DDMMSSToMilliarcsec(String value) instead. 
     */
    // This method is only used twice and with noCheck=true -- amchavan, 31-May-2005
    public static int DDMMSSToMilliarcsec(String value, boolean noCheck)
        throws IllegalArgumentException {

        return (stringToMilliarcsec(value, false, noCheck));
    }

    /**
     * Converts a String in the format +UU:MM:SS.sss to
     * milli-arcseconds, where + is an optional sign, UU is hours or
     * degrees, MM is minutes, SS is seconds and sss is millseconds.
     *
     * @param value	value to be converted, represented as a String.
     * @param ra	true means convert to HHMMSS, false to DDMMSS
     *
     * @return		input converted to a milliarcseconds
     *
     * @exception       IllegalArgumentException on out of bounds and
     *                                              out of format
     *
     * @author A M Chavan, 06-Jun-2001; was originally
     *         <code>convertToMilliArcSeconds()</code> in DegreeAdaptor.java
     */
    private static int stringToMilliarcsec(
        String value,
        boolean ra,
        boolean noCheck)
        throws IllegalArgumentException {

        // Drop useless leading "+" sign
        if (value.startsWith("+"))
            value = value.substring(1);

        // Let someone else deal with the no-colons case
        if (value.indexOf(':') == -1)
            return (stringToMilliarcsecNoColons(value, ra, noCheck));

        int result = 0;
        StringTokenizer st = new StringTokenizer(value, " :");

        try {
            int units;
            String unitsString = st.nextToken();
            int minutes = 0;
            int seconds = 0;
            int milliSecs = 0;
            int sign = +1;

            if (st.hasMoreTokens()) {
                minutes = new Integer(st.nextToken()).intValue();
                if (st.hasMoreTokens()) {
                    seconds = new Integer(st.nextToken(": ,.")).intValue();
                    if (st.hasMoreTokens()) {
                        String ms = st.nextToken(".,");
                        for (int j = 1; j <= DECIMAL_PLACES_ACCURACY; j++)
                            ms += "0";
                        ms = ms.substring(0, DECIMAL_PLACES_ACCURACY);
                        milliSecs = new Integer(ms).intValue();
                    }
                }
            }

            if (unitsString.startsWith("-")) {
                sign = -1;
                units = - new Integer(unitsString).intValue();
            } else {
                units = new Integer(unitsString).intValue();
            }

            if (!noCheck) {
                if (ra)
                    checkBoundsHHMMSS(sign, units, minutes, seconds, milliSecs);
                else
                    checkBoundsDDMMSS(sign, units, minutes, seconds, milliSecs);
            }

            result =
                sign
                    * (units * MILLISECS_PER_UNIT
                        + minutes * MILLISECS_PER_MINUTE
                        + seconds * MILLISECS_PER_SEC
                        + milliSecs);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException(e.getMessage());
        } catch (NoSuchElementException e) {
            throw new IllegalArgumentException(e.getMessage());
        }

        return result;
    }

    /**
     * Converts a String in the format [-][UUMMS]S[.sss] to
     * milli-arcseconds, where + is an optional sign, UU is hours or
     * degrees, MM is minutes, SS is seconds and sss is millseconds.
     *
     * @param value	value to be converted, represented as a String.
     * @param ra	true means convert to HHMMSS, false to DDMMSS
     * @param noCheck	do validation check
     *
     * @return		input converted to milliarcseconds
     *
     * @exception       IllegalArgumentException on out of bounds and
     *                                              out of format
     */
    private static int stringToMilliarcsecNoColons(
        String value,
        boolean ra,
        boolean noCheck)
        throws IllegalArgumentException {

        int result = 0; // return value
        int iPoint = -1; // index of decimal point
        int iSign = -1; // index of sign
        String main = null; // integer substring
        String frac = "000"; // fractional substring
        int t = 0; // scratch
        int l = 0; // scratch

        // Isolate 'integer' and 'fractional' 
        // parts of value string:
        // +123456.789 --> 123456 and 789
        //----------------------------------------

        value = value.trim();
        t = value.indexOf('.');
        l = value.length();
        iPoint = (t == -1) ? l : t;
        iSign = value.startsWith("-") ? 1 : 0;
        main = value.substring(iSign, iPoint).trim();
        if (iPoint < l) {
            // Input string has a decimal point:
            // we pad left with zeros, and that's it
            frac = value.substring(iPoint + 1, l) + "000";
            main = TextUtils.padded("000000", 6, main);
        } else {
            // Input string has no decimal point
            // We pad right with zeroes, but for the special
            // case of a single char value.
            // n    ->  0n0000 (same as 0n:00:00)
            // nm   ->  nm0000 (same as nm:00:00)
            // nmk  ->  nmk000 (same as nm:k0:00)
            // ... and so on.
            //
            // amchavan, 05-Oct-2001

            if (main.length() == 1)
                main = "0" + main;
            main = TextUtils.rightpadded("000000", 6, main);
        }
        frac = frac.substring(0, DECIMAL_PLACES_ACCURACY);

        if (main.length() > 6)
            throw new IllegalArgumentException(
                "String too long: \"" + value + "\"");
        String uu = main.substring(0, 2);
        String mm = main.substring(2, 4);
        String ss = main.substring(4, 6);

        try {

            int sign = (iSign == 1) ? -1 : 1;
            int units = new Integer(uu).intValue();
            int minutes = new Integer(mm).intValue();
            int seconds = new Integer(ss).intValue();
            int milliSecs = new Integer(frac).intValue();

            if (!noCheck) {
                if (ra)
                    checkBoundsHHMMSS(sign, units, minutes, seconds, milliSecs);
                else
                    checkBoundsDDMMSS(sign, units, minutes, seconds, milliSecs);
            }

            result =
                sign
                    * (units * MILLISECS_PER_UNIT
                        + minutes * MILLISECS_PER_MINUTE
                        + seconds * MILLISECS_PER_SEC
                        + milliSecs);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException(e.getMessage());
        } catch (NoSuchElementException e) {
            throw new IllegalArgumentException(e.getMessage());
        }

        return result;
    }

    /**
     * Converts milliarcseconds to a String in the format
     * "<code>&nbsp;HH:MM:SS.sss</code>" (note the leading space),
     * where HH is hours, MM is minutes, SS is seconds and sss is
     * millseconds.
     *
     * @param value	value to be converted, in milliarcseconds
     * @exception       IllegalArgumentException if value is negative
     *
     * @return		input converted to a String
     *
     * @author A M Chavan, 06-Jun-2001
     */
    public static String milliarcsecToHHMMSS(int value)
        throws IllegalArgumentException {
        return (milliarcsecToString(value, true, false));
    }

    /**
     * Converts seconds to a String in the format
     * "<code>&nbsp;HH:MM:SS.sss</code>" (note the leading space),
     * where HH is hours, MM is minutes, SS is seconds and sss is
     * milliseconds.
     *
     * @param value	value to be converted, in seconds
     * @exception	IllegalArgumentException if value is negative
     *
     * @return		input converted to a String
     *
     * @author cmavros, 10-10-01
     * @see #milliarcsecToHHMMSS(int)
     */
    public static String secToHHMMSS(int value)
        throws IllegalArgumentException {
        long millisecs = (long) value * (long) MILLISECS_PER_SEC;
        return milliarcsecToString(millisecs, false, true);
    }

    /**
     * Converts milliarcseconds to a String in the format
     * "<code>+DD:MM:SS.sss</code>", where + is an optional sign or
     * space, DD is hours, MM is minutes, SS is seconds and sss is
     * millseconds.
     *
     * @param value	value to be converted, in milliarcseconds
     *
     * @return		input converted to a String
     * @exception       IllegalArgumentException
     *
     * @author A M Chavan, 06-Jun-2001
     */
    public static String milliarcsecToDDMMSS(int value)
        throws IllegalArgumentException {
        return (milliarcsecToString(value, false, false));
    }

    /**
     * Converts milliarcseconds to a String in the format
     * +UU:MM:SS.sss, where + is an optional sign or space, UU is
     * hours or degrees, MM is minutes, SS is seconds and sss is
     * millseconds.
     *
     * @param value	value to be converted, in milliarcseconds
     * @param ra	true means convert to HH:MM:SS, 
     *                  false to DD:MM:SS
     *
     * @param noCheck	do validation check
     *
     * @return		input converted to a String
     *
     * @author A M Chavan, 06-Jun-2001; was originally
     *         convertToHMS() in DegreeAdaptor.java
     */
    private static String milliarcsecToString(
        long value,
        boolean ra,
        boolean noCheck)
        throws IllegalArgumentException {

        // Check input value
        if (!noCheck) {
            if (value > Integer.MAX_VALUE || value < Integer.MIN_VALUE) {
                throw new IllegalArgumentException(
                    "Integer value out of bounds: " + value);
            }
            if (ra) {
                if (!isValidRA((int) value)) {
                    throw new IllegalArgumentException(
                        "Invalid RA value: " + value);
                }
            } else if (!isValidDec((int) value)) {
                throw new IllegalArgumentException(
                    "Invalid DEC value: " + value);
            }
        }

        char sign = ' ';
        if (value < 0) {
            sign = '-';
            value = -value;
        }

        long units = value / MILLISECS_PER_UNIT;
        value -= units * MILLISECS_PER_UNIT;

        long minutes = value / MILLISECS_PER_MINUTE;
        value -= minutes * MILLISECS_PER_MINUTE;

        return sign
            + TextUtils.zeroPadded(2, units)
            + ":"
            + TextUtils.zeroPadded(2, minutes)
            + ":"
            + TextUtils.zeroPadded(2, (value / MILLISECS_PER_SEC))
            + "."
            + TextUtils.zeroPadded(3, (value % MILLISECS_PER_SEC));
    }

    /**
     * Check if an RA value in milliarcseconds is within its valid
     * range.
     *
     * @param ra       Right ascension in milliarcseconds.
     *
     * @return true if the value is within its range, false otherwise.
     *
     * @author A M Chavan, 04-Jul-2001
     */
    public static boolean isValidRA(int ra) {
        if (ra < 0 || ra >= MAX_MILLISECS_RA)
            return (false);
        return (true);
    }

    /**
     * Check if a Dec value in milliarcseconds is within its valid
     * range.
     *
     * @param dec       Declination in milliarcseconds.
     *
     * @return true if the value is within its range, false otherwise.
     *
     * @author A M Chavan, 04-Jul-2001
     */
    public static boolean isValidDec(int dec) {
        if (dec <=  -MAX_MILLISECS_DEC || dec > MAX_MILLISECS_DEC)
            return (false);
        return (true);
    }

    /**
     * Convert RA or Dec from milliarcseconds to hhmmss.sss, ie to
     * hours minutes and seconds plus fractional seconds.
     * 
     * @param n		right ascension or declination in 
     *                  milliarcseconds.<P>
     *                  <strong>Note</strong>: value is not checked.
     * 
     * @return		input argument converted to hhmmss.sss format
     *
     * @author A M Chavan, 07-Jun-2001: moved here from TextUtils class.
     * 
     * @see #milliarcsecToHHMMSS(int)
     * @see #milliarcsecToDDMMSS(int)
     * @deprecated Use #milliarcsecToHHMMSS(int) or #milliarcsecToDDMMSS(int)
     *             instead.
     */
    public static String convertToXXMMSS(int n) {
        String sign = "";

        if (n < 0) {
            sign = "-";
            n = -n;
        }

        int millisecs = n % 1000;
        n /= 1000;

        int secs = n % 60;
        n /= 60;

        int mins = n % 60;
        n /= 60;

        int xxx = n; // xxx is either hours or degrees.

        return sign
            + TextUtils.zeroPadded(2, xxx)
            + TextUtils.zeroPadded(2, mins)
            + TextUtils.zeroPadded(2, secs)
            + "."
            + TextUtils.zeroPadded(3, millisecs);
    }

    /**
     * Checks that the fields in a right ascension have legal values.
     * If any value is illegal, then it throws an exception.
     *
     * @param sign	must be +1
     * @param hours	value in the hours field of the RA
     * @param minutes	value in the minutes field of the RA
     * @param seconds	value in the seconds field of the RA
     * @param millisecs	value in the milliseconds field of the RA
     */
    private static void checkBoundsHHMMSS(
        int sign,
        int hours,
        int minutes,
        int seconds,
        int milliSecs)
        throws IllegalArgumentException {

        if (sign < 0
            || hours >= 24
            || minutes >= 60
            || seconds >= 60
            || milliSecs >= 1000) {
            throw new IllegalArgumentException(
                "Illegal HHMMSS value: "
                    + valuesToString(sign, hours, minutes, seconds, milliSecs));
        }
    }

    /**
     * Checks that the fields in a declination have legal values.
     * If any value is illegal it throws an exception.
     *
     * @param sign	-1 if the entry had a minus sign, +1 otherwise
     * @param degrees	value in the hours field of the RA
     * @param minutes	value in the minutes field of the RA
     * @param seconds	value in the seconds field of the RA
     * @param millisecs	value in the milliseconds field of the RA
     */
    private static void checkBoundsDDMMSS(
        int sign,
        int degrees,
        int minutes,
        int seconds,
        int milliSecs)
        throws IllegalArgumentException {
        if (degrees > 90
            || minutes >= 60
            || seconds >= 60
            || milliSecs >= 1000) {
            throw new IllegalArgumentException(
                "Illegal DDMMSS value: "
                    + valuesToString(sign, degrees, minutes, seconds, milliSecs));
        }
        if (degrees == 90 && (minutes > 0 || seconds > 0 || milliSecs > 0)) {
            throw new IllegalArgumentException(
                "Illegal DDMMSS value: "
                    + valuesToString(sign, degrees, minutes, seconds, milliSecs));
        }
    }

    /**
     * @return A colon separated representation of its input parameter.
     * @author A M Chavan, 16-Aug-2001: creation.
     */
    private static String valuesToString(
        int sign,
        int degrees,
        int minutes,
        int seconds,
        int milliSecs) {
        String ret = null;

        if (sign < 0)
            ret = "-";
        else
            ret = "";

        ret = ret + degrees + ":" + minutes + ":" + seconds + "." + milliSecs;
        return (ret);
    }

    /** 
     * Rounds argument db to a number of decimal places.
     * <BR>
     * Example: <code>round(2.32839, 1)</code> returns <code>2.3</code>
     *
     * @param db  - a double number
     * @param places -  the number of decimal places to round to
     *
     * @return double round to specific decimal places
     * 
     * @author cmavros, 19-Sept-2001 
     */
    public static double round(double db, int places) {
        double base = 1.;

        for (int i = 1; i <= places; i++) {
            base *= 10.;
        }
        double tmp = db * base;
        if (tmp >= 0.) {
            tmp += 0.5;
        } else
            tmp -= 0.5;
        int ip = (int) tmp;
        tmp = ((double) ip) / base;
        return (tmp);
    }

    /** 
     * Convert an ISO date+time string to seconds.
     *
     * The input string is parsed using a subset of the ISO
     * representation, and it is allowed to include less then the
     * entire ISO string; missing information is supplied by the
     * system based on the current time and date.
     * <P>
     *
     * <strong>Note</strong>: Input string is interpreted as UT.
     *
     * @param isodate  A possibly incomplete ISO date+time string
     *
     * @return The number of seconds since January 1, 1970,
     *         01:00:00 GMT represented by the input string.
     *
     * @throws ParseException
     *
     * @see #longToISODate( long )
     * @see #completeISODateString( String )
     *
     * @author amchavan, 09-Oct-2001: creation.
     */

    public static long ISOdateToLong(String isodate) throws ParseException {

        long milliSeconds = ISODateToMilliSec(isodate);
       long seconds = milliSeconds / 1000;
        return seconds;
    }

    /**
     * @param isodate
     * @return
     * @throws ParseException long
     */
    public static long ISODateToMilliSec(String isodate) throws ParseException {
        Date date = ISODateToDate(isodate);
       long milliSeconds = date.getTime();
        return milliSeconds;
    }

    /**
     * @see #ISOdateToLong( String )
     * @param isodate
     * @return
     * @throws ParseException Date
     */
    public static Date ISODateToDate(String isodate) throws ParseException {
        String fulldate = completeISODateString(isodate);
        SimpleDateFormat iso = new SimpleDateFormat(ISO_DATE_FORMAT);
        Date date = iso.parse(fulldate);
        return date;
    }

    /**
     * Convert an integer number of seconds to an ISO date+time
     * string. <BR>
     * All (absolute) times and dates are displayed in OHS applications
     * using the full ISO representation of date/time values,
     * which follows the following syntax:<BR>
     * &nbsp;&nbsp;&nbsp;&nbsp;<code>yyyy-mm-ddThh:mm:ss</code><bR>
     * where:
     * <BR>     
     *
     * <code>yyyy</code> is a four-digit year number, like
     * <code>1997</code>;<BR>
     *
     * <code>mm</code> is the two-digit month number, like
     * <code>08</code> for August;<BR>
     *
     * <code>dd</code> is the two digit day number, like
     * <code>14</code>;<BR>
     *
     * <code>hh</code> is the two digit hour number (24-hour clock),
     * like <code>21</code>;<BR>
     *
     * <code>mm</code> is the two digit minute number, like
     * <code>05</code>;<BR>
     *
     * <code>ss.sss</code> is the fractional number of seconds, like
     * <code>30.5</code>;<BR>
     * <P>
     *
     * <strong>Note</strong>: output string is expressed as UT.
     *
     * @author amchavan, 09-Oct-2001: creation.
     *
     * @param secs The number of seconds since January 1, 1970,
     *             01:00:00 GMT (Unix date) to be represented by
     *             the output string.
     *
     * @return An ISO date+time string. 
     *
     * @see #ISOdateToLong(String)
     */
    public static String longToISODate(long secs) {
        long milliSec = secs * 1000;      
        String isoDate = longToISODateMilliSec(milliSec);
        return isoDate;
    }

  
    
    /**
     * @param milliSec
     * @return String An ISO date+time string. 
     */
    public static String longToISODateMilliSec(long milliSec) {
          Date date = new Date(milliSec);
          String isoDate = dateToIsoDate(date);
        return isoDate;
    }

    /**
     * @param date
     * @return String An ISO date+time string. 
     */
    public static String dateToIsoDate(Date date) {
        SimpleDateFormat iso = new SimpleDateFormat(ISO_DATE_FORMAT);                 
        String isoDate = iso.format(date);
        return isoDate;
    }
    
    /**
     *
     * @param secs The number of seconds since January 1, 1970,
     *             01:00:00 GMT to be represented by the output 
     *             string.
     *
     * @return An ISO date+time string. 
     *
     * @see #ISOdateToLong(String)
     */
    public static String longToHHSS(long secs) {

        SimpleDateFormat iso = new SimpleDateFormat(ST_TIME_FORMAT);
        return (iso.format(new Date(secs * 1000)));
    }


    /**
     * Parses a partial date-time string. <P>
     *
     * The input string is parsed using a subset of the ISO
     * representation, and it is allowed to include less then the
     * entire ISO string; missing information is supplied by the
     * system: year and month are assumed to be the same as "today",
     * with hours, minutes and seconds being zero.
     * <br>
     * For more info, see <code>completeISODateString( String, String )</code>
     *
     * @param dt_in    Incomplete string to parse.
     *
     * @throws ParseException
     *
     * @see #completeISODateString( String, String )
     *
     * @author Toni Schlichting
     * @author amchavan, 09-Oct-2001: moved here from TIView.java,
     *                                and readapted.
     */
    public static String completeISODateString(String dt_in)
        throws ParseException {
        String now = TextUtils.currentISODateTime();
        int idx = now.indexOf('T');
        now = now.substring(0, idx + 1) + "00:00:00";
        return completeISODateString(dt_in, now);
    }

    /**
     * Parses a partial date-time string. <P>
     *
     * The input string is parsed using a subset of the ISO
     * representation, and it is allowed to include less then the
     * entire ISO string; missing information is supplied by parameter
     * <code>defDT</code>. The minimum piece of information to be
     * provided is the day number. The following table gives a few
     * examples of how partial ISO date/time strings are completed,
     * assuming <code>defDT</code> is <code>1997-08-01T00:00:00</code>.
     * <BR>
     * Notice that:
     * <UL>
     * <LI> blanks surrounding the partial strings are ignored, and
     *      are shown here only for clarity;     
     * <LI> one-digit numbers can, but need not, be prefixed by a zero
     *      ("2" is the same as "02"); 
     * <LI> the "T" char is only needed when a time different
     *      from midnight (00:00:00) needs to be specified; 
     * <LI> partial year specification (like "98") is not allowed;
     * <LI> textual month specification (like "Dec") is not allowed;
     * <LI> only years between 1970 and 2037 are supported.
     * </UL>
    <PRE>
            07              1997-08-07T00:00:00
         10-07              1997-10-07T00:00:00
    1998-2-1                1998-02-01T00:00:00
    1998-01-01              1998-01-01T00:00:00
            15T05           1997-08-15T05:00:00
            15T05:30        1997-08-15T05:30:00
            20T5:5:30       1997-08-20T05:05:30
         10-16T21:30        1997-10-16T21:30:00
    1999-12-31T23:59:59.99  1999-12-31T23:59:59.99
    98-2-1                  ERROR: partial year specification
    1969-12-31              ERROR: only years 1970-2037 are supported
    1999-12-31T23:59:61     ERROR: invalid number of seconds
    0001-55-55T66:66:66     ERROR: invalid time specification
    2001-Oct-09             ERROR: textual month specification is not allowed
    </PRE>
     *
     * @param inDT     Incomplete string to parse.
     * @param defDT    A full ISO date+time string; this is where
     *                 default information is taken from.<BR>
     *                 For instance: <code>2001-10-09T10:52:30</code><BR>
     *                 This parameter is not checked for integrity.
     *
     * @throws ParseException
     *
     * @see #completeISODateString( String )
     *
     * @author amchavan, 09-Oct-2001: creation
     */
    public static String completeISODateString(String inDT, String defDT)
        throws ParseException {

        String[] inElements = parseISODate(inDT);
        String[] defElements = parseISODate(defDT);
        StringBuffer complete = new StringBuffer();
        String ret = null;

        // Provide default elements
        //----------------------------------------
        for (int i = 0; i < inElements.length; i++)
            if (inElements[i].length() == 0)
                inElements[i] = defElements[i];

        // Build final string
        //----------------------------------------
        complete.append(TextUtils.zeroPadded(4, inElements[0]));
        complete.append('-');
        complete.append(TextUtils.zeroPadded(2, inElements[1]));
        complete.append('-');
        complete.append(TextUtils.zeroPadded(2, inElements[2]));
        complete.append('T');
        complete.append(TextUtils.zeroPadded(2, inElements[3]));
        complete.append(':');
        complete.append(TextUtils.zeroPadded(2, inElements[4]));
        complete.append(':');
        complete.append(TextUtils.zeroPadded(2, inElements[5]));

        ret = complete.toString();

        // Now verify if it makes any sense at all
        //----------------------------------------
        SimpleDateFormat iso = new SimpleDateFormat(ISO_DATE_FORMAT);
        iso.setLenient(false);
        Date d = iso.parse(ret);
        if (d == null)
            throw new ParseException(ret, 0);

        try {
            int y = Integer.parseInt(inElements[0]);
            if (y < 1970 || y > 2037)
                throw new ParseException(ret, 0);
        } catch (NumberFormatException e) {
            // should never happen here
        }

        return (ret); // All is OK, return
    }

    /**
     * Breaks up an ISO date+time string into an array of six
     * strings: year, month, day, hour, minutes and seconds.
     * 
     * @throws ParseException
     * @author amchavan, 09-Oct-2001: creation
     */
    private static String[] parseISODate(String isodatetime)
        throws ParseException {

        String[] components = new String[6];
        String t = null;
        String isodate = "";
        String isotime = "";

        isodatetime = isodatetime.trim(); // just in case

        // Check if we have a "T"
        int idx = isodatetime.indexOf('T');
        if (idx == -1)
            isodate = isodatetime; // No T, just date 
        else {
            isodate = isodatetime.substring(0, idx);
            isotime = isodatetime.substring(idx + 1);
        }

        // Parse the date part
        //----------------------------------------

        t = isodate;
        for (int i = 2; i >= 0; i--) {

            idx = t.lastIndexOf('-');
            if (idx == -1) {

                components[i] = t;
                t = "";
                continue;
            }

            components[i] = t.substring(idx + 1);
            t = t.substring(0, idx);
        }

        // Parse the time part
        //----------------------------------------

        t = isotime;
        for (int i = 3; i <= 5; i++) {

            idx = t.indexOf(':');
            if (idx == -1) {

                components[i] = t;
                t = "";
                continue;
            }

            components[i] = t.substring(0, idx);
            t = t.substring(idx + 1);
        }

        for (int i = 0; i < components.length; i++) {
            for (int j = 0; j < components[i].length(); j++) {

                char c = components[i].charAt(j);
                if (!Character.isDigit(c) && c != '.')
                    throw new ParseException(components[i], j);
            }
        }

        return (components);
    }
    
    /**
	 * Converts in milliseconds a String in the formats:
	 * 
	 * HH.sss
	 * HH MM.sss
	 * HH MM SS.sss
	 * (with spaces or : as separator)
	 * 
	 * 
	 * @param value
	 *            value to be converted, represented as a String.
	 * 
	 * 
	 * @return input converted to a milliarcseconds
	 * 
	 * @exception IllegalArgumentException
	 *                on out of bounds and out of format
	 * 
	 * <br>
	 * ChangeLog: fsogni, 02-Jun-2005; 
	 */
	
	public static int extendedRaToMilliarcsec (String ra) {
		
	return extendedStringToMilliarcsec (ra, true);
	
	}
	
	
	/**
	 * Converts in milliseconds a String in the formats:
	 * 
	 * DD.sss
	 * DD MM.sss
	 * DD MM SS.sss
	 * (with spaces or : as separator)
	 * 
	 * 
	 * @param value
	 *            value to be converted, represented as a String.
	 * 
	 * 
	 * @return input converted to a milliarcseconds
	 * 
	 * @exception IllegalArgumentException
	 *                on out of bounds and out of format
	 * 
	 * <br>
	 * ChangeLog: fsogni, 02-Jun-2005; 
	 */
	public static int extendedDecToMilliarcsec (String dec) {
	
	return extendedStringToMilliarcsec (dec, false);
	
	}
	
	
	// Converts a string from the format:
	// UU.sss
	// UU MM.sss
	// to the format UU MM SS.sss
	// and convert the value to milliseconds.
	private static int extendedStringToMilliarcsec(String value, boolean ra) {

		if (ra) {
		
		//    	 Rejects in case there is a sign: RA should not have signs.
		if (value.startsWith("+") || value.startsWith("-")) {
			throw new IllegalArgumentException(
					"By convention, Ra doesn't have a sign");
		}

		}
		
		
		int result = 0;
		StringTokenizer st = new StringTokenizer(value, " :");

		int units;

		int minutes = 0;
		int seconds = 0;
		int milliSecs = 0;
		int sign = +1;
		String hourString = st.nextToken();
		String minutesString = "00";
		String secondsString = "00";
		String milliSecsString = "000";

		StringTokenizer hoursTk = new StringTokenizer(hourString, ",.");
		if (hoursTk.hasMoreTokens()) {
			hourString = hoursTk.nextToken();
			if (hoursTk.hasMoreTokens()) {
				int hourFraction = new Integer(hoursTk.nextToken()).intValue();
				minutesString = String.valueOf(getMinutes(hourFraction));
				secondsString = String.valueOf(getSeconds(hourFraction));
				milliSecsString = String.valueOf(getMillisecondsFromHoursFraction(hourFraction));
			}
		} 
		
		if (st.hasMoreTokens()) {
		         minutesString = st.nextToken();
		         StringTokenizer minutesTk = new StringTokenizer(minutesString, " ,.");
		         minutesString = minutesTk.nextToken();
					if (minutesTk.hasMoreTokens()) {
						int minutesFraction = new Integer(minutesTk.nextToken()).intValue();
						secondsString = String.valueOf(getMinutes(minutesFraction));
						milliSecsString = String.valueOf(getMillisecondsFrominutesFraction(minutesFraction));
					}
		         	
					
		            if (st.hasMoreTokens()) {
		                seconds = new Integer(st.nextToken(": ,.")).intValue();
		                secondsString = String.valueOf(seconds);
		                if (st.hasMoreTokens()) {
		                    String ms = st.nextToken(".,");
		                    for (int j = 1; j <= DECIMAL_PLACES_ACCURACY; j++)
		                        ms += "0";
		                    ms = ms.substring(0, DECIMAL_PLACES_ACCURACY);
		                    milliSecs = new Integer(ms).intValue();
		                    milliSecsString = String.valueOf(milliSecs);
		                }
		            }
		        }

		String newRa = hourString + ":" + minutesString + ":" + secondsString + "." + milliSecsString;
		
		int milliarcsecs = stringToMilliarcsec(newRa, ra, false);
//		System.out.println("Hours: " + hourString);
//		System.out.println("Minutes: " + minutesString);
//		System.out.println("Seconds: " + secondsString);
//		System.out.println("Milliseconds: " + milliSecsString);
//		System.out.println("Milliarcseconds: " + milliarcsecs);
		return milliarcsecs;
	}

	//Extract the number of minutes from the integer.
	private static int getMinutes(int fraction) {

		int fract = fraction * 60 / 1000;
		return fract;
	}

	//Extract the number of seconds from the integer.
	private static int getSeconds(int fraction) {

		int modulus = (fraction*60) % 1000;
		int firstFraction = getMinutes(modulus);
		return firstFraction;
	}

	//Extract the number of milliseconds from the integer.
	private static int getMillisecondsFromHoursFraction(int fraction) {

		int firstFraction = (((fraction*60) % 1000)*60) % 1000;
		return firstFraction;
	}
	
//	Extract the number of milliseconds from the integer.
	private static int getMillisecondsFrominutesFraction(int fraction) {

		int firstFraction = (fraction*60) % 1000;
		return firstFraction;
	}
	
	
	
	
	/**
	 * Check if an RA value in the formats
	 * 
	 * HH.sss
	 * HH MM.sss
	 * HH MM SS.sss
	 * 
	 * is valid and within its valid range.
	 * 
	 * @param ra
	 *            Right ascension in a string format
	 * 
	 * @return true if the value is valid and within its range, false otherwise.
	 * 
	 * <br>
	 * ChangeLog: fsogni, 03-Jun-2005
	 */
	public static boolean isValidRA(String ra) {
	
		try {
			extendedRaToMilliarcsec(ra);
		} catch (IllegalArgumentException e) {
			return false;
		}
		return true;
	}
	
	


	
	/**
	 * Check if an DEC value in the formats
	 * 
	 * DD.sss
	 * DD MM.sss
	 * DD MM SS.sss
	 * 
	 * is valid and within its valid range.
	 * 
	 * @param dec
	 *            Declination in a string format
	 * 
	 * @return true if the value is valid and within its range, false otherwise.
	 * 
	 * <br>
	 * ChangeLog: fsogni, 03-Jun-2005
	 */
	public static boolean isValidDec(String dec) {
		
			try {
				extendedDecToMilliarcsec(dec);
			} catch (IllegalArgumentException e) {
				return false;
			}
			return true;
		}
	

    
    
    
    
    
    
}
