package org.eso.ohs.core.utilities;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

/**
 * Java Class for handling Fits checksum operations
 * <p>
 * Detailed information about the FITS file format and the associated
 * FITS checksum algorithm is available at:
 * <a href=http://archive.eso.org/fits-tools.html>http://archive.eso.org/fits-tools.html</a>
 * <p>
 * This class implements in Java the reference C algorithm for Fits checksum
 * computation.  Note that calculation of a Fits checksum is slower in Java
 * than in C for various reasons.
 * <p>
 * Note also that Java does not support unsigned types.
 * Since a 32 bit unsigned int type is not available, long (a signed 64 bit
 * wide type) must be used instead (and the top 32 bits should be ignored).
 * <p>
 * This utility may be invoked from the command line as follows:
 * <pre>
 *   java org.eso.ohs.core.utilities.FitsChecksum [filename]
 * </pre>
 * <p>
 * From within an application, use as follows:
 * <pre>
 *   long checksum = FitsChecksum.getChecksum(filename);
 *   String checksumString = FitsChecksum.getChecksumString(checksum);
 *
 *   long compliment = FitsChecksum.MAX32BIT - checksum;
 *   String complimentString = FitsChecksum.getChecksumString(compliment);
 * </pre>
 * @version $Revision: 1.2 $
 * @author Ron J Lobo
 */
public class FitsChecksum {

    /** cvs filename/version-number/timestamp of this class */
    public static final String cvsVersion_ =
        "$Id: FitsChecksum.java,v 1.2 2004/03/16 11:26:21 tcanavan Exp $";

    /** maximum value of a 32 bit unsigned number */
    public static final long MAX32BIT = 0xFFFFFFFFL;


    private static boolean verbose_ = false;
    private static boolean writeKey_ = false;
    private static boolean checkOnly_ = false;
    private static boolean inverse_ = false;

    private static String filename_ = null;
    private static String ascii_ = null;
    private static long numBytes_ = 0;

    // constants
    private static final int RECORD = 2880;
    private static final int BLOCK = 10;
    private static final int CHUNK = BLOCK*RECORD;


    /**
     * private constructor - this class should not be instantiated
     */
    private FitsChecksum() {}


    /**
     * This utility may be used as a standalone application.
     * It can be used to find out the Fits checksum of a file by invocation
     * from the command line  as follows:
     * <pre>
     *  java org.eso.ohs.core.utilities.FitsChecksum [filename]
     * </pre>
     */
    public static void main(String[] args) {

        parseCmdLineArgs(args);

        try {
            long chksum = getChecksum(filename_);
            long compli = MAX32BIT - chksum;

            // -x command line option
            if (checkOnly_) {
                System.exit(compli == 0 ? 0 : 1);
            }

            String checksumStr = getChecksumString(chksum);
            String compliStr = getChecksumString(compli);

            // -v command line option
            if (verbose_) {
                printChksum("checksum32", chksum, checksumStr);
                printChksum("compliment", compli, compliStr);
                System.out.println(" file size:  " + numBytes_ + " bytes");
                System.out.println();
            } else {
                System.out.println(checksumStr.substring(1,16) + checksumStr.charAt(0));
            }

        } catch (FileNotFoundException e) {
            System.err.println("Fits file not found: " + e.getMessage());
        } catch (IOException e) {
            System.err.println("An IO error has occurred: " + e.getMessage());
        }
    }


    /**
     * a method which returns the checksum of a Fits file given a filename
     *
     * @param filename  name of a file
     * @return checksum (64 bit value - the top 32 bits may be ignored)
     */
    public static long getChecksum(String filename)
    throws FileNotFoundException, IOException {
        return getChecksum(new FileInputStream(filename));
    }


    /**
     * a method which returns the FITS checksum of a string
     *
     * @param str       string to operate on
     * @param appendEOF whether to append an EOF character to the string
     * @return checksum (64 bit value - the top 32 bits may be ignored)
     */
    public static long getChecksum(String str, boolean appendEOF)
    throws IOException {
        if (appendEOF) {
            str += '\010';
        }
        return getChecksum(new ByteArrayInputStream(str.getBytes()));
    }


    /**
     * a method which returns the FITS checksum of an InputStream
     *
     * @param is        InputStream to operate on
     * @return checksum (64 bit value - the top 32 bits may be ignored)
     */
    public static long getChecksum(InputStream is) throws IOException {

        BufferedInputStream in = new BufferedInputStream(is, CHUNK);
        boolean ok = true;
        numBytes_ = 0;
        long sum32 = 0;
        int[] buf = new int[CHUNK];
        int i, b1;

        while (ok) {
            for (i=0; i < CHUNK; i++) {
                b1 = in.read();
                if (b1 == -1) {
                    ok = false;
                    break;
                } else {
                    buf[i] = b1;
                }
            }
            sum32 = getChunkChecksum(buf, i, sum32);
            numBytes_ += i;
        }
        in.close();
        return sum32;
    }


    private static long getChunkChecksum(int[] buf, int length, long sum32) {

        int len = 2 * (length / 4);
        int remain = length % 4;
        // System.out.println("length = " + length);
        // System.out.println("len    = " + len);
        // System.out.println("remain = " + remain + "\n");

        long hi = sum32 >> 16;
        long lo = sum32 & 0xFFFF;
        // printVal("hi: ", hi);
        // printVal("lo: ", lo);

        for (int i=0; i < len*2; i+=4) {
            hi += (buf[i] << 8) + buf[i+1];
            lo += (buf[i+2] << 8) + buf[i+3];
            // System.out.println("i="+i+", hi="+hi+", lo="+lo);
        }

        if (remain >= 1) {
            // System.out.println("remain=" + remain);
            hi += buf[2*len] * 0x100;
            if (remain >= 2) {
                hi += buf[2*len+1];
                if (remain >= 3) {
                    lo += buf[2*len+2] * 0x100;
                }
            }
        }

        long hicarry = hi >> 16;
        long locarry = lo >> 16;
        // printVal("hicarry: ", hicarry);
        // printVal("locarry: ", locarry);

        while (hicarry > 0 || locarry > 0) {
            // System.out.println("carry");
            hi = (hi & 0xFFFF) + locarry;
            lo = (lo & 0xFFFF) + hicarry;
            hicarry = hi >> 16;
            locarry = lo >> 16;
        }
        // printVal("hi: ", hi);
        // printVal("lo: ", lo);

        sum32 = (hi << 16) + lo;
        // printVal("sum32: ", sum32);
        return sum32;
    }


    /**
     * Encode an unsigned integer into a printable ASCII string.
     * The input bytes are each represented by four output bytes
     * whose sum is equal to the input integer, offset by 0x30 per byte.
     * The output is restricted to alphanumerics.
     * <p>
     * This is intended to be used to embed the complement of a file checksum
     * within an (originally 0'ed) ASCII field in the file.  The resulting
     * file checksum will then be the 1's complement -0 value (all 1's).
     * This is an additive identity value among other nifty properties.  The
     * embedded ASCII field must be 16 or 32 bit aligned, or the characters
     * can be permuted to compensate.
     * <p>
     * To invert the encoding, simply subtract the offset from each byte
     * and pass the resulting string to checksum.
     * <p>
     * @param sum       a 32 bit checksum value (top 32 bits should be empty)
     * @return String   encoded 16 character ASCII checksum string
     */
    public static String getChecksumString(long sum) {

        long value = MAX32BIT - (sum & MAX32BIT);
        int offset = 0x30;     /* ASCII 0 (zero) */
        long mask[] = { 0xff000000L, 0xff0000L, 0xff00L, 0xffL };
        int exclude[] = { 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40,
                          0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60 };
        int[] ch = new int[4];
        char[] asc = new char[32];

        for (int i=0; i < 4; i++) {

            long bite = (value & mask[i]) >> (24 - (8 * i));
            long quotient = bite / 4 + offset;
            long remainder = bite % 4;
            for (int j=0; j < 4; j++) {
                ch[j] = (int)quotient;
            }
            ch[0] += remainder;

            int check;
            do {
                check = 0;
                for (int k=0; k < 13; k++) {
                    for (int j=0; j < 4; j += 2) {
                        if (ch[j] == exclude[k] || ch[j+1] == exclude[k]) {
                            ch[j]++;
                            ch[j+1]--;
                            check++;
                        }
                    }
                }
            } while (check != 0);

            for (int j=0; j < 4; j++) {
                asc[j*4+i] = (char) ch[j];
            }
        }

        String ascii = "";
        for (int i=0; i < 16; i++) {
            ascii += (char)asc[(i+15)%16];
        }
        return ascii;
    }


    private static void parseCmdLineArgs(String[] args) {

        // set variables to defaults
        verbose_ = false;
        writeKey_ = false;
        checkOnly_ = false;
        inverse_ = false;

        filename_ = null;
        ascii_ = null;

        boolean gotName = false;

        // parse cmd-line options
        for (int i = 0; i < args.length; i++) {

            if (args[i].startsWith("-")) {

                if (args[i].equals("-v")) {
                    verbose_ = true;
                    continue;
/*
                } else if (args[i].equals("-w")) {
                    writeKey_ = true;
                    continue;
*/
                } else if (args[i].equals("-x")) {
                    checkOnly_ = true;
                    continue;
/*
                } else if (args[i].equals("-i")) {
                    if (++i >= args.length) {
                        printUsage();
                        System.exit(1);
                    } else {
                        ascii_ = args[i];
                        inverse_ = true;
                        continue;
                    }
*/
                } else {
                    System.err.println("error: unknown argument: " + args[i]);
                    printUsage();
                    System.exit(1);
                }

            } else if (!gotName) {
                filename_ = args[i];
                gotName = true;
                continue;

            } else {
                System.err.println("error: too many arguments: " + args[i]);
                printUsage();
                System.exit(1);
            }
        }

        if (!gotName) {
            System.err.println("error: no filename given");
            printUsage();
            System.exit(1);
        }
    }


    private static void printUsage () {
        System.out.println("Usage: " +
        "fits-checksum [-v] [-w] [-x] [-i <ascii>] <file>\n" +
        "Command line flags:\n" +
        "\t-v      verbose output, print 32 bit checksum and compliment\n" +
        "\t-i      invert the transformation given an ascii string\n" +
        "\t-x      check a file whose checksum has already been zeroed \n" +
        "\t        exits with 0 if ok otherwise with 1\n" +
        "\t-w      overwrite CHECKSUM keyword in FITS header if present\n\n");
    }


    /**
     * method to neatly display a value in decimal and hex
     */
    protected static void printVal(String name, long l) {
        System.out.println(name + l + " = " + Long.toHexString(l));
    }


    /**
     * method to neatly display a checksum
     */
    protected static void printChksum(String name, long val, String chkStr) {
        String valStr = "0000000000" + val;
        int len = valStr.length();
        valStr = valStr.substring(len-10, len);
        System.out.println(name + ":  " + valStr + " = " + chkStr);
    }
}

