package org.eso.ohs.dfs;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Vector;

import org.eso.ohs.core.docview.datatrans.DocViewObject;

/**
 * Common superclass for all business objects.
 * 
 * <p>
 * $Id: BusinessObject.java,v 1.37 2005/04/21 13:27:54 tcanavan Exp $
 * <br>
 * @author Bob Kemp
 */

public abstract class BusinessObject extends SimpleStorableObject
implements DocViewObject, Comparable {

    /** Version id for object serialization */
    static final long serialVersionUID = -7436117420493246341L;
	
    /**
     * What is the id of the object from which this one was cloned?
     * Used for database check-in/out.
     * 0 if we're not a clone.
     */
    private long cloneParentId_ = 0;

    /**
     * User id that created the object.
     * This is needed when the object is modified in operations mode,
     * and hence needs to be saved under a different user id (the
     * original owner's id) than * the one currently active.
     */
    private int ownerId_;

    /**
     * has the object been checked-in?
     * In other words, is it a local copy of a checked-in object?
     * It is an integer because Targets, ODS, etc can be checked-in
     * multiple times and we can only make them read/write after
     * every checked-in version has been checked-out.
     */
    private int checkedInCount_ = 0;

    /**
     *  When the object was created -- in Unix time
     */
    private long creationDate_;

    /**
     *  When the object was last modified -- in Unix time
     */
    private long lastModifiedDate_;

    /**
     *  Id of object in the database or 0 if never checked-in
     */
    private long dbaseId_;

    /**
     * used by the receiver. It represents the line number of the .tex file.
     * 
     */
    private int lineNumber_ = 0;

    /**
     * Default constructor.
     * Initialise property change event support.
     */
    public BusinessObject() {
		// Sets Last Modification Date as side-effect
		// (like all setX() methods, except setLastModifiedDate())
	setCreationDate(timeNowInSeconds());
    }

    private long timeNowInSeconds() {
	return System.currentTimeMillis()/1000;
    }

    /**
     * Mark this BO children as being already checked in.
     */
    public void markCheckedIn() {
	checkedInCount_++;	// another copy checked-in
	firePropertyChangeEvent();
    }


    /**
     * Mark this BO and all children as being already checked in.
     */
    public void recursiveMarkCheckedIn() {
	BusinessObject[] bos = getBusObjChildren();

	markCheckedIn();

	for (int j=0; j < bos.length; j++) {
	    bos[j].recursiveMarkCheckedIn();
	}
    }

    /**
     * Decrement checkedInCount_ property
     */
    public void markCheckedOut() {
	if (checkedInCount_ > 0) {
	    checkedInCount_--;	// a copy checked-out
	}
	else {
	    checkedInCount_ = 0;
	}
	firePropertyChangeEvent();
    }

    /**
     * Decrement checkedInCount_ property
     */
    public void recursiveMarkCheckedOut() {
	BusinessObject[] bos = getBusObjChildren();

	markCheckedOut();

	for (int j=0; j < bos.length; j++) {
	    bos[j].recursiveMarkCheckedOut();
	}
    }

    /**
     * Get value for checkedInCount_ property
     *
     * @return the current value of the property
     */
    public boolean isCheckedIn() {
	return checkedInCount_ != 0;
    }

    /**
     * Set the id of the checked-in equivalent of this object.
     * 
     * @param dbaseId	id of the checked-in version of this BusinessObject
     */
    public void setDbaseId(long dbaseId) {
	dbaseId_ = dbaseId;
	firePropertyChangeEvent();
    }

    /**
     * Get the id of the checked-in equivalent of this object.
     * 
     * @return id of the checked-in version of this BusinessObject
     */
    public long getDbaseId() {
	return dbaseId_;
    }

    /**
     * Set the owner id for this BO
     * 
     * @param ownerId	the id of the owner of this BusinessObject
     */
    public void setOwnerId(int ownerId) {
	ownerId_ = ownerId;
	firePropertyChangeEvent();
    }

    /**
     * Get the owner id for this BO
     * 
     * @return the id of the owner of this BusinessObject
     */
    public int getOwnerId() {
	return ownerId_;
    }

    /**
     * Get value for cloneParentId_ property
     *
     * @return the current value of the property
     */
    public long getCloneParentId() {
	return cloneParentId_;
    }

    /**
     * Set new value for cloneParentId_ property
     *
     * @param newValue the new value for the property
     */
    public void setCloneParentId(long newValue) {
	cloneParentId_ = newValue;
	firePropertyChangeEvent();
    }

    /**
     * Set new value for creationDate_ property
     *
     * @param newValue the new value for the property
     */
    public void setCreationDate(long newValue) {
	creationDate_ = newValue;
	firePropertyChangeEvent();
    }

    /**
     * Get value for creationDate_ property
     *
     * @return the current value of the property
     */
    public long getCreationDate() {
	return creationDate_;
    }


    /**
     * Set new value for lastModifiedDate_ property
     *
     * @param newValue the new value for the property
     */
    public void setLastModifiedDate(long newValue) {
	lastModifiedDate_ = newValue;
	    // NOTE: We do not call firePropertyChangeEvent() here
	    //       because that would reset the modification time.
    }

    /**
     * Get value for lastModifiedDate_ property
     *
     * @return the current value of the property
     */
    public long getLastModifiedDate() {
	return lastModifiedDate_;
    }

    /**
     * Verifies this block and all attached blocks, returning the
     * concatenation of all error messages.
     * 
     * @return the (possibly empty) concatenation of all error messages.
     */
    public String verify() {
	return "";
    }


    /**
     * Verifies this block and all attached blocks, returning the
     * concatenation of all error messages.
     * 
     * @return the (possibly empty) concatenation of all error messages.
     */
    public String verifyBO() {
	return "";
    }



    /**
     * Returns an array of the immediate child BusinessObjects for
     * this BO.  
     * 
     * @return an array of business objects referred to by this BO.
     */
    public BusinessObject[] getBusObjChildren() {
	return new BusinessObject[] {};
    }


    /**
     * Returns a string representation of the object.
     * Mainly used for debugging.
     *
     * @return the string representation
     */
    public String toString() {
	String className = getClass().getName();

	return className.substring(className.lastIndexOf('.')+1)
		+ "[" + paramString() + "]";
    }

    /**
     * Returns a deep copy of this BO, ie recursively copies this BO
     * and all of it's attached BOs.
     * 
     * @return a deep copy of this business object.
     */
    public BusinessObject deepCopy() {
	return copy();
    }

    /**
     * Create the list of property names for (subclass of) BusinessObject.
     * 
     * <p>X is a property of type T for an object, if there is a public
     * getX() method returning an object of type T and a public setX(T)
     * method that takes a single parameter of type T.
     * 
     * @return a list of property names.
     */
    public String[] propertyNames() {
	return propertyNames(false);
    }

    /**
     * Create the list of property names for (subclass of) BusinessObject,
     * excluding the id and those referring to BusinessObjects.
     * 
     * <p>X is a property of type T for an object, if there is a public
     * getX() method returning an object of type T and a public setX(T)
     * method that takes a single parameter of type T.
     * 
     * @return a list of property names.
     */
    public String[] simpleProperties() {
	return propertyNames(true);
    }

    /**
     * Create the list of property names for (subclass of) BusinessObject,
     * including or excluding the id and those referring to BusinessObjects.
     * 
     * <p>X is a property of type T for an object, if there is a public
     * getX() method returning an object of type T and a public setX(T)
     * method that takes a single parameter of type T.
     * 
     * @param onlySimpleProps	true if properties referring to the id and
     *			    	attached (subclasses of) BusinessObject should
     *			    	be excluded, false otherwise.
     * 
     * @return a list of property names.
     */
    private String[] propertyNames(boolean simpleOnly) {
	Vector propNames = new Vector();
	Method[] meths = this.getClass().getMethods();

	for(int i=0;i<meths.length;i++) {
	    String name  = meths[i].getName();
	    Class retType = meths[i].getReturnType();

	    if(name.startsWith("get")) {
		String propertyName = name.substring(3);
		String setName = "set" + propertyName;

		if (simpleOnly) {
		    if (propertyName.equals("Id")) {
			continue;	// 
		    }
		    if (propertyName.equals("DbaseId")) {
			continue;	// 
		    }
		    if (BusinessObject.class.isAssignableFrom(retType)) {
			continue;	// 
		    }
		}

		for(int j=0;j<meths.length;j++) {
		    Method meth = meths[j];
		    Class[] parms = meth.getParameterTypes();

		    if(setName.equals(meth.getName())
			   && parms.length == 1
			   && parms[0].equals(retType)) {

			propNames.addElement(propertyName);
		    }
		}
	    }
	}
	String[] names = new String[propNames.size()];
	propNames.copyInto(names);
	return names;
    }


    /**
     * Copies the values of the specified properties of this object to
     * the given BusinessObject.
     * 
     * @param propNames	    array of property names to copy
     * @param bo	    target of the copy operation
     * 
     * @exception NoSuchMethodException
     * 			    The given property did not have a getXXX()
     * 			    and setXXX() pair of methods, ie it's not
     * 			    really a Javabeans property.
     * 
     * @exception IllegalAccessException
     * 			    One or both of the getXXX() and setXXX()
     * 			    methods is not public.
     * 
     * @exception InvocationTargetException
     * 			    The call to getXXX() or to setXXX()
     * 			    generated an exception.  The original
     * 			    exception is accessible from this one
     * 			    using its getTargetException() method.
     */
    public void copyProperty(String propName, BusinessObject bo)
			throws NoSuchMethodException,
				IllegalAccessException,
				InvocationTargetException {
	Class cls = bo.getClass();
	Method get = cls.getMethod(
			    "get" + propName,
			    new Class[] { }
		       );
	Method set = cls.getMethod(
			    "set" + propName,
			    new Class[] {get.getReturnType()}
		       );                

	Object obj = get.invoke(this, new Object[] {});
	set.invoke(bo, new Object[] {obj});
    }


    /**
     * Copies the values of the specified properties of this object to
     * the given BusinessObject.
     * 
     * @param propNames	    array of property names to copy
     * @param bo	    target of the copy operation
     * 
     * @exception NoSuchMethodException
     * 			    The given property did not have a getXXX()
     * 			    and setXXX() pair of methods, ie it's not
     * 			    really a Javabeans property.
     * 
     * @exception IllegalAccessException
     * 			    One or both of the getXXX() and setXXX()
     * 			    methods is not public.
     * 
     * @exception InvocationTargetException
     * 			    The call to getXXX() or to setXXX()
     * 			    generated an exception.  The original
     * 			    exception is accessible from this one
     * 			    using its getTargetException() method.
     */
    public void copyProperties(String[] propNames, BusinessObject destBO)
			throws NoSuchMethodException,
				IllegalAccessException,
				InvocationTargetException {
        try {
            destBO.suspendPropertyChangeEvents();
	    for (int j=0; j < propNames.length; j++) {
	        copyProperty(propNames[j], destBO);
	    }
        }
        finally {
            destBO.resumePropertyChangeEvents();
            destBO.firePropertyChangeEvent();
        }
    }


    /**
     * Returns a copy of the current object, identical except for
     * it's id and the medium and directory containing it.
     * The id of the new object is set to zero but registering it
     * with the object manager will cause a new id to be assigned.
     * 
     * <p>
     * NOTE: This method does not register the object with the ObjectManager.
     * Within P2PP, you will probably want to use Persistence.copy().
     * 
     * @param device	the device on which the new object should reside
     * @param dir	the directory in which the new object should reside
     * @param baseId	Id of associated OB, or 0
     * 
     * @return		a clone of the current BusinessObject (or subclass)
     * 
     * @see org.eso.ohs.phase2.apps.p2pp.Persistence
     */
    public BusinessObject copy()
				    {
	Class cls = this.getClass();
	String[] propNames = this.propertyNames();
	BusinessObject bo = null;

	try {
	    bo = (BusinessObject)cls.newInstance();
	    copyProperties(propNames, bo);
	    bo.setId(0);

				// Remember where it was cloned from!
	    bo.setCloneParentId(getId());
	    bo.setDbaseId(0);
	    bo.setCreationDate(timeNowInSeconds());

	    while (bo.isCheckedIn()) {
		bo.markCheckedOut();
	    }

	    return bo;
	}
	catch (IllegalAccessException e) {
	    e.printStackTrace();
	}
	catch (InvocationTargetException e) {
	    e.getTargetException().printStackTrace();
	}
	catch (NoSuchMethodException e) {
	    e.printStackTrace();
	}
	catch (InstantiationException e) {
	    e.printStackTrace();
	}

	return null;
    }

    /**
     * Informs listeners of a property change.
     * No details of which property of the old/new value are given.
     */
    protected void firePropertyChangeEvent() {
	lastModifiedDate_ = timeNowInSeconds();
	super.firePropertyChangeEvent();
    }

    protected void firePropertyChangeEvent(PropertyChangeEvent ev) {
	lastModifiedDate_ = timeNowInSeconds();
	super.firePropertyChangeEvent(ev);
    }


    /**
     * Indirection method used by the Visitor design pattern.
     * <p>
     * The code for the method body should be
     * <ul>
     * return visitor.visit(this, arg)
     * </ul>
     * Where visitor has many overloaded definitions for visit().
     * 
     * The reason for this contortion is that the visitor does not know
     * exactly the type of this business object.  However, this object
     * knows its own type and will thus call the correct visit() method
     * on visitor.
     * 
     * @param visitor	object that is to act upon this business object.
     * @param arg	extra data that might be needed for the visitor
     * 
     * @return		result of calling the visitor.visit() method.
     * 
     * @see BusinessVisitor
     */
    public abstract Object accept(BusinessVisitor visitor, Object arg);

    /**
     * Hashcode for this class, dependent on some properties.
     *
     * <p>
     * Must be kept in sync with equals()
     * 
     * @return		hash value based basic properties of this object
     * 
     * @see #equals
     */
    public int hashCode() {
	return getOwnerId() ^
		getName().hashCode() ^
		(int)getId();
    }

    /**
     * Test object for equality, based on a few properties.
     * Used in unit tests.  Strictly speaking P2PP could
     * just use the Id property, but for unit tests, it's good
     * to use a few more properties.
     *
     * <p>
     * Must be kept in sync with hashCode()
     * 
     * @param obj	object to be tested for equality with this one.
     * 
     * @return		true if some basic properties are the same,
     * 			false otherwise
     * 
     * @see #hashCode
     */
    public boolean equals(Object obj) {
	if (obj instanceof BusinessObject) {
	    BusinessObject bo = (BusinessObject)obj;
	    return bo.getId() == getId()
		    && bo.getName().equals(getName())
		    && bo.getOwnerId() == getOwnerId();
	}
	return false;
    }

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object obj) {
		if (obj instanceof BusinessObject) {
		BusinessObject bo = (BusinessObject)obj;
		 return bo.getName().compareTo(getName());
	}
		return 0;
	}
    

    /**
     * @param lineNumber_ The lineNumber_ to set.
     */
    public void setLineNumber(int lineNumber_) {
        this.lineNumber_ = lineNumber_;
    }

    /**
     * @return Returns the lineNumber_.
     */
    public int getLineNumber() {
        return lineNumber_;
    }


    /**
     * Forwarder for PropertyChange events.
     * <p>
     * This is a bit of a hack but is useful in the ForderGridView
     * where we are displaying the names of attached objects as if they
     * were part of (eg) the OB.  Thus when the attached object's
     * name changes, we need this BO to appear to have changed.
     */
    public static class EventForwarder implements PropertyChangeListener,
					    Serializable {

	/** Version id for object serialization */
	static final long serialVersionUID = -4315758539348021046L;

	/** BO to be notified of changes*/
	private BusinessObject bo_;

	public EventForwarder(BusinessObject bo) {
	    bo_ = bo;
	}

	public void propertyChange(PropertyChangeEvent ev) {       
	  // Hack I think the event forwarder can be removed as we no longer
          // share networks.
	    if(ev.getPropertyName() != null)
	        bo_.firePropertyChangeEvent(ev);
	    else
	        bo_.firePropertyChangeEvent();
	}
    }

}


