package org.eso.ohs.instruments;

import java.io.IOException;
import java.util.List;

import org.apache.log4j.Logger;

/**
 * A background thread to initialise the list of instruments.
 * It is only used at system startup, while the cache, etc is
 * being read.
 *
 * <p>
 * $Id: InstrumentInitialiser.java,v 1.19 2005/11/17 12:26:18 tcanavan Exp $
 * <br>
 * @author Bob Kemp
 */

public class InstrumentInitialiser extends Thread {

    /** Standard log channel */
    private static Logger stdlog_ =
        Logger.getLogger(InstrumentInitialiser.class);

    /** Result of this subcomputation */
    InstrumentList result = null;

    /** Have we finished yet?  A synchronization variable */
    private boolean finished = false;

    /** The configured instrument package directory pathname 
    <br>ChangeLog: A M Chavan, 05-Jul-2000 */
    private String instDir_ = "";
    
    private String instCacheDir_="";
    
    private boolean ipMode_=false;
    
    /** 
     * List of error encountered during building the InstrumentList
     */
    private StringBuffer errorList_ = new StringBuffer();
    
    
    public InstrumentInitialiser(boolean ipMode,String instDir,String instCacheDir) {
        instDir_ = instDir;
        ipMode_ = ipMode;
        instCacheDir_ = instCacheDir;
    }
    

    /**
     * Assigns an instance of the instrument list to the
     * result variable thus causing the instrument files
     * to be read in from disk.
     */
    public synchronized void run() {

        try {
            result = InstrumentList.setupInstrumentList(ipMode_,instDir_,instCacheDir_);
        } catch (InstrumentException e1) {
            errorList_.append(e1.getMessage()+"\n");
            stdlog_.debug(e1.getMessage());
        } finally {
            finished = true;
            notifyAll();
        }
    }

    /**
     * Synchronization method for the parent thread (and any others).
     * This waits (if necessary) until the subcomputation is finished
     * and then returns the result of the subcomputation
     */
    public synchronized InstrumentList getList() {
        while (!finished) {
            try {
                wait();
            } catch (InterruptedException e) {
            }
        }
        return result;
    }

    /** 
     * Checks that the instrument files have been successfully loaded
     * and issues any error messages to the user.  Returns true if there
     * were no fatal errors.
     * 
     * @return	    true if there were no fatal errors, false otherwise
     */
    public boolean checkInstrumentsLoaded() throws IOException {
        InstrumentList instrList = getList();
        Instrument[] instList = null;
        if (instrList != null) {
            instList = instrList.getInstrumentList();
        }

        if (errorList_.length() > 0) {
            return false;
        }
        if (instList == null) {
            return false;
        } else {
            StringBuffer buf = new StringBuffer();
            for (int i = 0; i < instList.length; i++) {
                if (instList[i].isIndexOutOfDate())
                    buf.append(instList[i].getInstrumentName() + "\n");
            }
            if (buf.length() > 0) {
                buf.insert(
                    0,
                    "Insufficient access rights to create index file(s) for \n");
                buf.append("Using read only versions of index files(s)\n");
                throw new IOException(buf.toString());
            }
            return true;
        }
    }
    
    /** 
     * Checks that the instrument files have the correct version number.
     * 
     * @return	    true if there were no fatal errors, false otherwise
     */
    public boolean checkInstrumentsVersion() throws IOException {
        InstrumentList instrList = getList();        
        if (instrList != null) {
            List listIgnored = instrList.getIgnoredInstruments();
            StringBuffer buf = new StringBuffer();
            for(int i=0;i<listIgnored.size();i++) {
               buf.append(listIgnored.get(i).toString() + "\n");	
            }
            if(buf.length() > 0 ) {
            	buf.insert(0,"The following instruments have been assigned an invalid version number. ");
            	throw new IOException(buf.toString());
            }
        }
        return true;
    }
}
